from __future__ import annotations

import responses
from responses import matchers

from yandex_query_client import YQHttpClient, YQHttpClientConfig

IAM_TOKEN = "my_iam_token"
PROJECT = "my_project"


class TestYQHttpClient:
    def setup_method(self):
        config = YQHttpClientConfig(IAM_TOKEN, PROJECT)
        self.client = YQHttpClient(config)

    def setup_mocks_for_query_execution(self, query_results_json_list):
        responses.post(
            "https://api.yandex-query.cloud.yandex.net/api/fq/v1/queries",
            match=[
                matchers.header_matcher(
                    {"Content-Type": "application/json", "Authorization": f"Bearer {IAM_TOKEN}"}
                ),
                matchers.query_param_matcher({"project": PROJECT}),
                matchers.json_params_matcher({"name": "my query", "text": "select 777"}),
            ],
            json={"id": "query1"},
            status=200,
        )

        responses.get(
            "https://api.yandex-query.cloud.yandex.net/api/fq/v1/queries/query1/status",
            json={"status": "RUNNING"},
            status=200,
        )

        responses.get(
            "https://api.yandex-query.cloud.yandex.net/api/fq/v1/queries/query1/status",
            json={"status": "COMPLETED"},
            status=200,
        )

        result_set_count = len(query_results_json_list)
        responses.get(
            "https://api.yandex-query.cloud.yandex.net/api/fq/v1/queries/query1",
            json={
                "id": "query1",
                "result_sets": [{"rows_count": 1, "truncated": False} for _ in range(result_set_count)],
            },
            status=200,
        )

        for i in range(result_set_count):
            responses.get(
                f"https://api.yandex-query.cloud.yandex.net/api/fq/v1/queries/query1/results/{i}",
                json=query_results_json_list[i],
                status=200,
            )

    def _create_test_query(self):
        query_id = self.client.create_query(query_text="select 777", name="my query")
        assert query_id == "query1"
        return query_id

    @responses.activate()
    def test_select_results(self):
        self.setup_mocks_for_query_execution(
            [{"rows": [[777]], "columns": [{"name": "column0", "type": "Int32"}]}]
        )

        query_id = self._create_test_query()

        result_set_count = self.client.wait_query_to_succeed(query_id)
        assert result_set_count == 1
        results = self.client.get_query_all_result_sets(query_id, result_set_count=result_set_count)
        assert results == {"rows": [[777]], "columns": [{"name": "column0", "type": "Int32"}]}

        responses.post(
            "https://api.yandex-query.cloud.yandex.net/api/fq/v1/queries/query1/stop",
            match=[
                matchers.header_matcher({"Authorization": f"Bearer {IAM_TOKEN}"}),
                matchers.query_param_matcher({"project": PROJECT}),
            ],
            status=204,
        )

        assert self.client.get_query_status(query_id) == "COMPLETED"
        assert self.client.get_query(query_id) == {
            "id": "query1",
            "result_sets": [{"rows_count": 1, "truncated": False}],
        }
        self.client.stop_query(query_id)

        assert (
            self.client.compose_query_web_link(query_id)
            == f"https://yq.cloud.yandex.ru/folders/{PROJECT}/ide/queries/query1"
        )

    @responses.activate()
    def test_select_two_record_sets(self):
        self.setup_mocks_for_query_execution(
            [
                {"rows": [[777]], "columns": [{"name": "column0", "type": "Int32"}]},
                {"rows": [["zzz"]], "columns": [{"name": "aaaa", "type": "Utf8"}]},
            ]
        )

        query_id = self._create_test_query()

        result_set_count = self.client.wait_query_to_succeed(query_id)
        assert result_set_count == 2
        results = self.client.get_query_all_result_sets(query_id, result_set_count=result_set_count)
        assert results[0] == {"rows": [[777]], "columns": [{"name": "column0", "type": "Int32"}]}
        assert results[1] == {"rows": [["zzz"]], "columns": [{"name": "aaaa", "type": "Utf8"}]}



from __future__ import annotations

import base64
import pprint
from datetime import datetime
from decimal import Decimal
from typing import Any

import dateutil.parser


class YQResults:
    """Holds and formats query execution results."""

    def __init__(self, results: dict[str, Any]):
        self._raw_results = results
        self._results = None

    @staticmethod
    def _convert_from_float(value: float | str) -> float:
        # special values, e.g inf encoded as str, normal values are in float
        return float(value)

    @staticmethod
    def _convert_from_pgfloat(value: str | None) -> float | None:
        if value is None:
            return None
        return float(value)

    @staticmethod
    def _convert_from_pgint(value: str | None) -> int | None:
        if value is None:
            return None
        return int(value)

    @staticmethod
    def _convert_from_decimal(value: str) -> Decimal:
        return Decimal(value)

    @staticmethod
    def _convert_from_pgnumeric(value: str | None) -> Decimal | None:
        if value is None:
            return None
        return Decimal(value)

    @staticmethod
    def _convert_from_base64(value: str) -> str | bytes:
        b = base64.b64decode(value)
        try:
            return b.decode("utf-8")
        except UnicodeDecodeError:
            return b

    @staticmethod
    def _convert_from_datetime(value: str) -> datetime:
        # suitable for yql data and datetime parsing
        return dateutil.parser.isoparse(value)

    @staticmethod
    def _convert_from_pgdatetime(value: str | None) -> datetime | None:
        if value is None:
            return None
        return dateutil.parser.isoparse(value)

    @staticmethod
    def _convert_from_enum(value: list) -> str:
        return str(value[0])

    @staticmethod
    def _extract_from_optional(type_name: str) -> str:
        # Uint16? -> Uint16
        if type_name.endswith("?"):
            return type_name[0:-1]

        # Optional<Uint16> -> Uint16
        return type_name[len("Optional<") : -1]

    @staticmethod
    def _extract_from_set(type_name: str) -> str:
        # Set<Uint16> -> Uint16
        return type_name[len("Set<") : -1]

    @staticmethod
    def _extract_from_list(type_name: str) -> str:
        # List<Uint16> -> Uint16
        return type_name[len("List<") : -1]

    @staticmethod
    def _split_type_list(type_list: str) -> list[str]:
        # naive implementation
        # fixme: fix it
        return type_list.split(",")

    @staticmethod
    def _extract_from_tuple(type_name: str) -> list[str]:
        # Tuple<Uint16, String, Double> -> [Uint16, String, Double]
        return YQResults._split_type_list(type_name[len("Tuple<") : -1])

    @staticmethod
    def _extract_from_dict(type_name: str) -> tuple[str, str]:
        # Dict<Uint16, String> -> (Uint16, String)
        [key_type, value_type] = YQResults._split_type_list(type_name[len("Dict<") : -1])
        return key_type, value_type

    @staticmethod
    def _extract_from_variant_over_struct(type_name: str) -> dict[str, str]:
        # Variant<'One':Int32,'Two':String> -> {One: Int32, Two: String}
        types_with_names = YQResults._split_type_list(type_name[len("Variant<") : -1])
        result = {}
        for t in types_with_names:
            [n, t] = t.split(":")
            # strip '
            n = n[1:-1]
            result[n] = t
        return result

    @staticmethod
    def _extract_from_variant_over_tuple(type_name: str) -> list[str]:
        # Variant<Int32,String> -> [Int32, String]
        return YQResults._split_type_list(type_name[len("Variant<") : -1])

    @staticmethod
    def _convert_from_optional(value: list[Any]) -> Any:
        # Optional types are encoded as [[]] objects
        # If type is Uint16, value is encoded as {"rows":[[value]]}
        # If type is Optional<Uint16>, value is encoded as {"rows":[[[value]]]}
        # If value is None than result is {"rows":[[[]]]}
        # So check if len equals 1 it means that it contains value
        # if len is 0 it means it has no value i.e. value is None
        assert len(value) < 2, str(value)
        if len(value) == 1:
            return value[0]

        return None

    @staticmethod
    def id(v):
        return v

    @staticmethod
    def _get_converter(column_type: str) -> Any:
        """Return converter based on column type."""
        # primitives
        if column_type in [
            "Int8",
            "Int16",
            "Int32",
            "Int64",
            "Uint8",
            "Uint16",
            "Uint32",
            "Uint64",
            "Bool",
            "Utf8",
            "Uuid",
            "Void",
            "Null",
            "EmptyList",
            "Struct<>",
            "Tuple<>",
        ]:
            return YQResults.id

        if column_type == "String":
            return YQResults._convert_from_base64

        if column_type in ["Float", "Double"]:
            return YQResults._convert_from_float

        if column_type.startswith("Decimal("):
            return YQResults._convert_from_decimal

        if column_type.startswith("Enum<"):
            return YQResults._convert_from_enum

        if column_type in ["Date", "Datetime", "Timestamp"]:
            return YQResults._convert_from_datetime

        # containers
        if column_type.startswith("Optional<") or column_type.endswith("?"):
            # If type is Optional than get base type
            inner_converter = YQResults._get_converter(YQResults._extract_from_optional(column_type))

            # Remove "Optional" encoding
            # and convert resulting value as others
            def convert(x):
                inner_value = YQResults._convert_from_optional(x)
                if inner_value is None:
                    return None
                return inner_converter(inner_value)

            return convert

        if column_type.startswith("Set<"):
            inner_converter = YQResults._get_converter(YQResults._extract_from_set(column_type))

            def convert(x):
                return {inner_converter(v) for v in x}

            return convert

        if column_type.startswith("List<"):
            inner_converter = YQResults._get_converter(YQResults._extract_from_list(column_type))

            def convert(x):
                return [inner_converter(v) for v in x]

            return convert

        if column_type.startswith("Tuple<"):
            inner_types_list = YQResults._extract_from_tuple(column_type)
            inner_converters_list = [YQResults._get_converter(t) for t in inner_types_list]

            def convert(x):
                assert len(x) == len(
                    inner_converters_list
                ), f"Wrong length for tuple value: {len(x)} != {len(inner_converters_list)}"
                return tuple([c(v) for (c, v) in zip(inner_converters_list, x)])

            return convert

        # variant over struct
        if column_type.startswith("Variant<'"):
            inner_types_dict = YQResults._extract_from_variant_over_struct(column_type)
            inner_converters_dict = {k: YQResults._get_converter(t) for k, t in inner_types_dict.items()}

            def convert(x):
                return inner_converters_dict[x[0]](x[1])

            return convert

        # variant over tuple
        if column_type.startswith("Variant<"):
            inner_types_list = YQResults._extract_from_variant_over_tuple(column_type)
            inner_converters_list = [YQResults._get_converter(t) for t in inner_types_list]

            def convert(x):
                return inner_converters_list[x[0]](x[1])

            return convert

        if column_type == "EmptyDict":

            def convert(x):
                return {}

            return convert

        if column_type.startswith("Dict<"):
            key_type, value_type = YQResults._extract_from_dict(column_type)
            key_converter = YQResults._get_converter(key_type)
            value_converter = YQResults._get_converter(value_type)

            def convert(x):
                return {key_converter(v[0]): value_converter(v[1]) for v in x}

            return convert

        # pg types
        if column_type.startswith("pgfloat"):
            return YQResults._convert_from_pgfloat

        if column_type in ["pgint2", "pgint4", "pgint8"]:
            return YQResults._convert_from_pgint

        if column_type == "pgnumeric":
            return YQResults._convert_from_pgnumeric

        if column_type in ["pgdate", "pgtimestamp"]:
            return YQResults._convert_from_pgdatetime

        if column_type.startswith("pg"):
            return YQResults.id

        # unsupported type
        return YQResults.id

    def _convert(self):
        converters = []
        converted_results = []
        for column in self._raw_results["columns"]:
            converters.append(YQResults._get_converter(column["type"]))

        for row in self._raw_results["rows"]:
            new_row = []
            for index, value in enumerate(row):
                converter = converters[index]
                new_row.append(value if converter is None else converter(value))

            converted_results.append(new_row)

        self._results = {"rows": converted_results, "columns": self._raw_results["columns"]}

    def _repr_pretty_(self, p, cycle):
        p.text(pprint.pformat(self._results))

    @property
    def results(self):
        if self._results is None:
            self._convert()

        return self._results

    @property
    def raw_results(self):
        return self._raw_results

    def to_table(self):
        return self._results["rows"]

    def to_dataframe(self):
        result_set = self._results
        columns = [column["name"] for column in result_set["columns"]]
        import pandas

        return pandas.DataFrame(result_set["rows"], columns=columns)



from __future__ import annotations

from datetime import datetime
from decimal import Decimal

from dateutil.tz import tzutc

from yandex_query_client import YQResults


class TestYQResults:
    def test_integral_results(self):
        # json response and results could be found here: https://github.com/ydb-platform/ydb/blob/284b7efb67edcdade0b12c849b7fad40739ad62b/ydb/tests/fq/http_api/test_http_api.py#L336
        r = YQResults(
            {
                "rows": [
                    [
                        100,
                        -100,
                        200,
                        200,
                        10000000000,
                        -20000000000,
                        "18014398509481984",
                        "-18014398509481984",
                        123.5,
                        -789.125,
                        "inf",
                        True,
                        False,
                        "aGVsbG8=",
                        "hello",
                        "1.23",
                        'he"llo_again',
                        "Я Привет",
                        1,
                        2,
                        3,
                        4,
                    ]
                ],
                "columns": [
                    {"name": "column0", "type": "Int32"},
                    {"name": "column1", "type": "Int32"},
                    {"name": "column2", "type": "Int64"},
                    {"name": "column3", "type": "Uint64"},
                    {"name": "column4", "type": "Uint64"},
                    {"name": "column5", "type": "Int64"},
                    {"name": "column6", "type": "Int64"},
                    {"name": "column7", "type": "Int64"},
                    {"name": "column8", "type": "Float"},
                    {"name": "column9", "type": "Double"},
                    {"name": "column10", "type": "Double"},
                    {"name": "column11", "type": "Bool"},
                    {"name": "column12", "type": "Bool"},
                    {"name": "column13", "type": "String"},
                    {"name": "column14", "type": "Utf8"},
                    {"name": "column15", "type": "Decimal(6,3)"},
                    {"name": "column16", "type": "Utf8"},
                    {"name": "column17", "type": "Utf8"},
                    {"name": "column18", "type": "Int8"},
                    {"name": "column19", "type": "Int16"},
                    {"name": "column20", "type": "Uint8"},
                    {"name": "column21", "type": "Uint16"},
                ],
            },
        )

        results = r.results

        assert results == {
            "rows": [
                [
                    100,
                    -100,
                    200,
                    200,
                    10000000000,
                    -20000000000,
                    "18014398509481984",
                    "-18014398509481984",
                    123.5,
                    -789.125,
                    float("inf"),
                    True,
                    False,
                    "hello",
                    "hello",
                    Decimal("1.23"),
                    'he"llo_again',
                    "Я Привет",
                    1,
                    2,
                    3,
                    4,
                ]
            ],
            "columns": [
                {"name": "column0", "type": "Int32"},
                {"name": "column1", "type": "Int32"},
                {"name": "column2", "type": "Int64"},
                {"name": "column3", "type": "Uint64"},
                {"name": "column4", "type": "Uint64"},
                {"name": "column5", "type": "Int64"},
                {"name": "column6", "type": "Int64"},
                {"name": "column7", "type": "Int64"},
                {"name": "column8", "type": "Float"},
                {"name": "column9", "type": "Double"},
                {"name": "column10", "type": "Double"},
                {"name": "column11", "type": "Bool"},
                {"name": "column12", "type": "Bool"},
                {"name": "column13", "type": "String"},
                {"name": "column14", "type": "Utf8"},
                {"name": "column15", "type": "Decimal(6,3)"},
                {"name": "column16", "type": "Utf8"},
                {"name": "column17", "type": "Utf8"},
                {"name": "column18", "type": "Int8"},
                {"name": "column19", "type": "Int16"},
                {"name": "column20", "type": "Uint8"},
                {"name": "column21", "type": "Uint16"},
            ],
        }

    def test_complex_results(self):
        # json response and results could be found here: https://github.com/ydb-platform/ydb/blob/284b7efb67edcdade0b12c849b7fad40739ad62b/ydb/tests/fq/http_api/test_http_api.py#L445
        r = YQResults(
            {
                "rows": [
                    [
                        [],
                        [1, 2],
                        [],
                        [["YWJj", 1]],
                        [["xyz", 1]],
                        None,
                        "PT15M",
                        "2019-09-16",
                        "2019-09-16T10:46:05Z",
                        "2019-09-16T11:27:44.345849Z",
                        "2019-09-16,Europe/Moscow",
                        "2019-09-16T14:32:40,Europe/Moscow",
                        "2019-09-16T14:32:55.874913,Europe/Moscow",
                        ["One", 12],
                        [1, "eHl6"],
                        ["a", 1],
                        ["monday", None],
                        1,
                        {},
                        {"a": 1, "b": "xyz"},
                        None,
                        None,
                        [[[1, [[177]]]]],
                        [[[1, []]]],
                        [[[1, []]]],
                        ["Foo", None],
                        ["Bar", None],
                        [],
                        [1, "cHJpdmV0", "2019-09-16"],
                    ]
                ],
                "columns": [
                    {"name": "column0", "type": "EmptyList"},
                    {"name": "column1", "type": "List<Int32>"},
                    {"name": "column2", "type": "EmptyDict"},
                    {"name": "column3", "type": "Dict<String,Int32>"},
                    {"name": "column4", "type": "Dict<Utf8,Int32>"},
                    {"name": "column5", "type": "Uuid"},
                    {"name": "column6", "type": "Interval"},
                    {"name": "column7", "type": "Date"},
                    {"name": "column8", "type": "Datetime"},
                    {"name": "column9", "type": "Timestamp"},
                    {"name": "column10", "type": "TzDate"},
                    {"name": "column11", "type": "TzDatetime"},
                    {"name": "column12", "type": "TzTimestamp"},
                    {"name": "column13", "type": "Variant<'One':Int32,'Two':String>"},
                    {"name": "column14", "type": "Variant<Int32,String>"},
                    {"name": "column15", "type": "Variant<'a':Int32>"},
                    {"name": "column16", "type": "Enum<'monday'>"},
                    {"name": "column17", "type": "Tagged<Int32,'my_tag'>"},
                    {"name": "column18", "type": "Struct<>"},
                    {"name": "column19", "type": "Struct<'a':Int32,'b':Utf8>"},
                    {"name": "column20", "type": "Void"},
                    {"name": "column21", "type": "Null"},
                    {"name": "column22", "type": "Optional<Variant<String,Int32??>?>"},
                    {"name": "column23", "type": "Optional<Variant<String,Int32??>?>"},
                    {"name": "column24", "type": "Optional<Variant<String,Int32?>?>"},
                    {"name": "column25", "type": "Enum<'Bar','Foo'>"},
                    {"name": "column26", "type": "Enum<'Bar','Foo'>"},
                    {"name": "column27", "type": "Tuple<>"},
                    {"name": "column28", "type": "Tuple<Int32,String,Date>"},
                ],
            },
        )

        results = r.results

        assert results == {
            "rows": [
                [
                    [],
                    [1, 2],
                    {},
                    {"abc": 1},
                    {"xyz": 1},
                    None,  # seems like http api doesn't support uuid values
                    "PT15M",
                    datetime(2019, 9, 16, 0, 0),
                    datetime(2019, 9, 16, 10, 46, 5, tzinfo=tzutc()),
                    datetime(2019, 9, 16, 11, 27, 44, 345849, tzinfo=tzutc()),
                    "2019-09-16,Europe/Moscow",
                    "2019-09-16T14:32:40,Europe/Moscow",
                    "2019-09-16T14:32:55.874913,Europe/Moscow",
                    12,
                    "xyz",
                    1,
                    "monday",
                    1,
                    {},
                    {"a": 1, "b": "xyz"},
                    None,
                    None,
                    177,
                    None,
                    None,
                    "Foo",
                    "Bar",
                    [],
                    (1, "privet", datetime(2019, 9, 16, 0, 0)),
                ]
            ],
            "columns": [
                {"name": "column0", "type": "EmptyList"},
                {"name": "column1", "type": "List<Int32>"},
                {"name": "column2", "type": "EmptyDict"},
                {"name": "column3", "type": "Dict<String,Int32>"},
                {"name": "column4", "type": "Dict<Utf8,Int32>"},
                {"name": "column5", "type": "Uuid"},
                {"name": "column6", "type": "Interval"},
                {"name": "column7", "type": "Date"},
                {"name": "column8", "type": "Datetime"},
                {"name": "column9", "type": "Timestamp"},
                {"name": "column10", "type": "TzDate"},
                {"name": "column11", "type": "TzDatetime"},
                {"name": "column12", "type": "TzTimestamp"},
                {"name": "column13", "type": "Variant<'One':Int32,'Two':String>"},
                {"name": "column14", "type": "Variant<Int32,String>"},
                {"name": "column15", "type": "Variant<'a':Int32>"},
                {"name": "column16", "type": "Enum<'monday'>"},
                {"name": "column17", "type": "Tagged<Int32,'my_tag'>"},
                {"name": "column18", "type": "Struct<>"},
                {"name": "column19", "type": "Struct<'a':Int32,'b':Utf8>"},
                {"name": "column20", "type": "Void"},
                {"name": "column21", "type": "Null"},
                {"name": "column22", "type": "Optional<Variant<String,Int32??>?>"},
                {"name": "column23", "type": "Optional<Variant<String,Int32??>?>"},
                {"name": "column24", "type": "Optional<Variant<String,Int32?>?>"},
                {"name": "column25", "type": "Enum<'Bar','Foo'>"},
                {"name": "column26", "type": "Enum<'Bar','Foo'>"},
                {"name": "column27", "type": "Tuple<>"},
                {"name": "column28", "type": "Tuple<Int32,String,Date>"},
            ],
        }



from __future__ import annotations

import logging
import time
from datetime import datetime
from typing import Any

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .query_results import YQResults

MAX_RETRY_FOR_SESSION = 4
BACK_OFF_FACTOR = 0.3
TIME_BETWEEN_RETRIES = 1000
ERROR_CODES = (500, 502, 504)

logger = logging.getLogger(__name__)


def _requests_retry_session(
    session, retries=MAX_RETRY_FOR_SESSION, back_off_factor=BACK_OFF_FACTOR, status_force_list=ERROR_CODES
):
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=back_off_factor,
        status_forcelist=status_force_list,
        allowed_methods=frozenset(["GET", "POST"]),
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


class YQHttpClientConfig:
    """YandexQuery HTTP client config."""

    def __init__(
        self,
        token: str | None = None,
        project: str | None = None,
        user_agent: str | None = "Python YQ HTTP SDK",
    ) -> None:
        self.token = token
        self.project = project
        self.user_agent = user_agent

        # urls should not contain trailing /
        self.endpoint: str = "https://api.yandex-query.cloud.yandex.net"
        self.web_base_url: str = "https://yq.cloud.yandex.ru"
        self.token_prefix = "Bearer "


class YQHttpClientException(Exception):
    """YandexQuery client exception type."""

    def __init__(
        self, message: str, status: str | None = None, msg: str | None = None, details: Any = None
    ) -> None:
        super().__init__(message)
        self.status = status
        self.msg = msg
        self.details = details


class YQHttpClient:
    """YandexQuery HTTP client."""

    def __init__(self, config: YQHttpClientConfig):
        self.config = config
        self.session = _requests_retry_session(session=requests.Session())

    def __enter__(self):
        """Return the object when a context manager is created."""
        return self

    def __exit__(self, *args):
        """Close network connection when exiting the context manager."""
        self.session.close()

    def close(self):
        """Close network connection."""
        self.session.close()

    def _build_headers(self, idempotency_key=None, request_id=None) -> dict[str, str]:
        headers = {"Authorization": f"{self.config.token_prefix}{self.config.token}"}
        if idempotency_key is not None:
            headers["Idempotency-Key"] = idempotency_key

        if request_id is not None:
            headers["x-request-id"] = request_id

        if self.config.user_agent is not None:
            headers["User-Agent"] = self.config.user_agent

        return headers

    def _build_params(self) -> dict[str, str]:
        params = {}
        if self.config.project is not None:
            params["project"] = self.config.project

        return params

    def _compose_api_url(self, path: str) -> str:
        return self.config.endpoint + path

    def _compose_web_url(self, path: str) -> str:
        return self.config.web_base_url + path

    def _validate_http_error(self, response, expected_code=200) -> None:
        logger.debug("Response: %s, %s", response.status_code, response.text)
        if response.status_code != expected_code:
            if response.headers.get("Content-Type", "").startswith("application/json"):
                body = response.json()
                status = body.get("status")
                msg = body.get("message")
                details = body.get("details")
                raise YQHttpClientException(
                    f"Error occurred. http code={response.status_code}, status={status}, msg={msg}, details={details}",
                    status=status,
                    msg=msg,
                    details=details,
                )

            raise YQHttpClientException(f"Error occurred: {response.status_code}, {response.text}")

    def create_query(
        self,
        query_text=None,
        query_type=None,
        name=None,
        description=None,
        idempotency_key=None,
        request_id=None,
        expected_code=200,
    ):
        body = {}
        if query_text is not None:
            body["text"] = query_text

        if query_type is not None:
            body["type"] = query_type

        if name is not None:
            body["name"] = name

        if description is not None:
            body["description"] = description

        response = self.session.post(
            self._compose_api_url("/api/fq/v1/queries"),
            headers=self._build_headers(idempotency_key=idempotency_key, request_id=request_id),
            params=self._build_params(),
            json=body,
        )

        self._validate_http_error(response, expected_code=expected_code)
        return response.json()["id"]

    def get_query_status(self, query_id, request_id=None, expected_code=200) -> Any:
        response = self.session.get(
            self._compose_api_url(f"/api/fq/v1/queries/{query_id}/status"),
            headers=self._build_headers(request_id=request_id),
            params=self._build_params(),
        )

        self._validate_http_error(response, expected_code=expected_code)
        return response.json()["status"]

    def get_query(self, query_id, request_id=None, expected_code=200) -> Any:
        response = self.session.get(
            self._compose_api_url(f"/api/fq/v1/queries/{query_id}"),
            headers=self._build_headers(request_id=request_id),
            params=self._build_params(),
        )

        self._validate_http_error(response, expected_code=expected_code)
        return response.json()

    def stop_query(
        self,
        query_id: str,
        idempotency_key: str | None = None,
        request_id: str | None = None,
        expected_code: int = 204,
    ) -> Any:
        headers = self._build_headers(idempotency_key=idempotency_key, request_id=request_id)
        response = self.session.post(
            self._compose_api_url(f"/api/fq/v1/queries/{query_id}/stop"),
            headers=headers,
            params=self._build_params(),
        )
        self._validate_http_error(response, expected_code=expected_code)
        return response

    def wait_query_to_complete(self, query_id, execution_timeout=None, stop_on_timeout=False) -> str:
        status = None
        delay = 0.2  # start with 0.2 sec
        try:
            start = datetime.now()
            while True:
                if execution_timeout is not None and datetime.now() > start + execution_timeout:
                    raise TimeoutError(f"Query {query_id} execution timeout, last status {status}")

                status = self.get_query_status(query_id)
                if status not in ["RUNNING", "PENDING"]:
                    return status

                time.sleep(delay)
                delay *= 2
                delay = min(2, delay)  # up to 2 seconds

        except TimeoutError:
            if stop_on_timeout:
                self.stop_query(query_id)
            raise

    def wait_query_to_succeed(self, query_id, execution_timeout=None, stop_on_timeout=False) -> int:
        status = self.wait_query_to_complete(
            query_id=query_id, execution_timeout=execution_timeout, stop_on_timeout=stop_on_timeout
        )

        query = self.get_query(query_id)
        if status != "COMPLETED":
            issues = query["issues"]
            raise RuntimeError(f"Query {query_id} failed with issues={issues}")

        return len(query["result_sets"])

    def get_query_result_set_page(
        self,
        query_id,
        result_set_index,
        offset=None,
        limit=None,
        raw_format=False,
        request_id=None,
        expected_code=200,
    ) -> Any:
        params = self._build_params()
        if offset is not None:
            params["offset"] = offset

        if limit is not None:
            params["limit"] = limit

        response = self.session.get(
            self._compose_api_url(f"/api/fq/v1/queries/{query_id}/results/{result_set_index}"),
            headers=self._build_headers(request_id=request_id),
            params=params,
        )

        self._validate_http_error(response, expected_code=expected_code)
        return response.json()

    def get_query_result_set(self, query_id: str, result_set_index: int, raw_format: bool = False) -> Any:
        offset = 0
        limit = 1000
        columns = None
        rows = []
        while True:
            part = self.get_query_result_set_page(
                query_id, result_set_index=result_set_index, offset=offset, limit=limit, raw_format=raw_format
            )

            if columns is None:
                columns = part["columns"]

            r = part["rows"]
            rows.extend(r)
            if len(r) != limit:
                break

            offset += limit

        result = {"rows": rows, "columns": columns}
        if raw_format:
            return result

        return YQResults(result).results

    def get_query_all_result_sets(
        self, query_id: str, result_set_count: int, raw_format: bool = False
    ) -> Any:
        result = []
        for i in range(0, result_set_count):
            r = self.get_query_result_set(query_id, result_set_index=i, raw_format=raw_format)

            if result_set_count == 1:
                return r

            result.append(r)
        return result

    def get_openapi_spec(self) -> str:
        response = self.session.get(self._compose_api_url("/resources/v1/openapi.yaml"))
        self._validate_http_error(response)
        return response.text

    def compose_query_web_link(self, query_id) -> str:
        return self._compose_web_url(f"/folders/{self.config.project}/ide/queries/{query_id}")

    @staticmethod
    def result_set_to_dataframe(data):
        import pandas as pd

        column_names = [column["name"] for column in data["columns"]]
        return pd.DataFrame(data["rows"], columns=column_names)
